from typing import Callable, Optional, List, Union, Tuple, Iterable, Any

import numpy as np
import pandas as pd
import tensorflow as tf

from mmproteo.utils.utils import unzip


class SequenceEvaluator:
    def __init__(
            self,
            dataset: tf.data.Dataset,
            decode_func: Optional[Callable[[np.ndarray], np.ndarray]],
            batch_size: int = 32,
            separator: str = " ",
            padding_character: str = "_",
            prediction_col_name: str = "predicted",
            true_value_col_name: str = "true",
    ):
        self.dataset = dataset
        self.decode_func = decode_func
        self.batch_size = batch_size
        self.separator = separator
        self.padding_character = padding_character
        self.prediction_col_name = prediction_col_name
        self.true_value_col_name = true_value_col_name

    def evaluate_model(
            self,
            model: tf.keras.Model,
            steps: Optional[int] = 0,
    ) -> Union[float, List[float]]:
        if steps is None:
            return model.evaluate(self.dataset)

        if steps == 0:
            steps = int(40000 / self.batch_size)

        return model.evaluate(self.dataset.repeat(), steps=steps)

    def _shorten_sequences_to_lengths_of_other_sequences_with_separator(
            self,
            sequences: pd.Series,
            other_sequences: pd.Series,
            length_offset: int = 0,
            split_by_separator: bool = True,
    ) -> pd.Series:

        if split_by_separator:
            other_sequences = other_sequences.str.rstrip(
                self.padding_character + self.separator
            )
            other_sequences = other_sequences.str.split(self.separator)

            def shorten_sequence(
                    sequence: str,
                    length: int
            ) -> str:
                return self.separator.join(
                    sequence.split(self.separator)[:length]
                )
        else:
            other_sequences = other_sequences.str.rstrip(self.padding_character)

            def shorten_sequence(
                    sequence: str,
                    length: int
            ) -> str:
                return sequence[:length]

        lengths = other_sequences.str.len()
        lengths += length_offset

        return sequences.combine(
            other=lengths,
            func=shorten_sequence,
        )

    def evaluate_model_visually(
            self,
            model: tf.keras.Model,
            sample_size: int = 20,
            keep_separator: bool = True,
            split_by_separator: bool = True,
    ) -> Tuple[pd.DataFrame, Tuple[Iterable, Iterable, Any]]:
        eval_ds = self.dataset.unbatch().batch(1).take(sample_size)
        x_eval: Tuple[np.ndarray]
        y_eval: Tuple[np.ndarray]
        x_eval, y_eval = unzip(eval_ds.as_numpy_iterator())  # type: ignore
        y_pred: np.ndarray = model.predict(eval_ds)

        eval_df = pd.DataFrame(
            data=zip(
                self._decode_indices_to_str(
                    array=y_pred,
                    onehot=True,
                    flatten=True,
                ),
                self._decode_indices_to_str(
                    array=y_eval,
                    onehot=False,
                    flatten=True,
                )
            ),
            columns=[self.prediction_col_name, self.true_value_col_name]
        )

        eval_df[self.prediction_col_name] = \
            self._shorten_sequences_to_lengths_of_other_sequences_with_separator(
                sequences=eval_df[self.prediction_col_name],
                other_sequences=eval_df[self.true_value_col_name],
                length_offset=1,
                split_by_separator=split_by_separator,
            )

        if not keep_separator:
            eval_df = eval_df.applymap(lambda s: s.replace(self.separator, ""))

        eval_df[self.true_value_col_name] = eval_df[self.true_value_col_name]\
            .str.rstrip(
                self.padding_character + self.separator
            )

        return eval_df, (x_eval, y_eval, y_pred)

    @staticmethod
    def _decode_onehot_to_index(array: np.ndarray) -> np.ndarray:
        return np.argmax(array, axis=-1)

    def _concat_letter_rows_with_separator(self, array: np.ndarray) \
            -> np.ndarray:
        return np.apply_along_axis(lambda row: self.separator.join(row),
                                   axis=-1, arr=array)

    def _decode_indices_to_str(
            self,
            array: Union[np.ndarray, Iterable[np.ndarray]],
            onehot: bool = True,
            flatten: bool = True,
    ) -> np.ndarray:
        if onehot:
            array = self._decode_onehot_to_index(array)

        if self.decode_func is not None:
            array = self.decode_func(array)

        array = self._concat_letter_rows_with_separator(array)

        if flatten:
            array = array.squeeze()

        return array



import gc
import os
import shutil
import time
from typing import List, Dict, Union, Callable, Any, Optional, Iterable, Tuple

import numpy as np
import pandas as pd
import tensorflow as tf
import tensorflow_datasets as tfds

from mmproteo.utils import log, utils
from mmproteo.utils.processing import ItemProcessor


class Parquet2DatasetFileProcessor:

    def __init__(self,
                 training_data_columns: List[str],
                 target_data_columns: List[str],
                 padding_lengths: Dict[str, int],
                 padding_characters: Dict[str, Union[str, int, float]],
                 column_normalizations: Dict[str, Callable[[Any], Any]],
                 dataset_dump_path_prefix: str,
                 char_to_idx_mappers: Optional[Dict[str, Dict[str, int]]] = None,
                 char_to_idx_mapping_functions: Optional[Dict[str, Callable[[str], int]]] = None,
                 item_count: int = 0,
                 skip_existing: bool = True,
                 split_on_column_values_of: Optional[List[str]] = None,
                 logger: log.Logger = log.DEFAULT_LOGGER
                 ):
        self.training_data_columns = training_data_columns
        self.target_data_columns = target_data_columns
        self.padding_lengths = padding_lengths
        self.padding_characters = padding_characters
        self.column_normalizations = column_normalizations
        self.char_to_idx_mapping_functions = char_to_idx_mapping_functions
        if self.char_to_idx_mapping_functions is None:
            assert char_to_idx_mappers is not None, \
                "either char_to_idx_mappers or char_to_idx_mapping_functions must be given"
            self.char_to_idx_mapping_functions = {
                column: mapping.get for column, mapping in char_to_idx_mappers.items()  # type: ignore
            }
        self.item_count = item_count
        self.dataset_dump_path_prefix = dataset_dump_path_prefix
        self.char_idx_dtype = np.int8
        self.skip_existing = skip_existing
        self.split_on_column_values_of = split_on_column_values_of
        self.logger = logger

    def normalize_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.column_normalizations is None:
            return df
        df = df.copy()
        for column, normalize_func in self.column_normalizations.items():
            df[column] = df[column].apply(normalize_func)
        return df

    def pad_array_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        if len(df) == 0:
            return df

        df = df.copy()
        for column, padding_length in self.padding_lengths.items():
            item_dtype = df[column].iloc[0].dtype

            df[column] = list(tf.keras.preprocessing.sequence.pad_sequences(
                sequences=df[column],
                maxlen=padding_length,
                padding='post',
                value=self.padding_characters[column],
                dtype=item_dtype
            ))
        return df

    @staticmethod
    def _sequence_to_indices(sequence: Iterable[str],
                             char_to_idx_mapping_func: Callable[[str], int],
                             dtype: type) -> np.ndarray:
        return np.array([char_to_idx_mapping_func(char) for char in sequence],
                        dtype=dtype)

    def sequence_column_to_indices(self, df: pd.DataFrame) -> pd.DataFrame:
        assert self.char_to_idx_mapping_functions is not None, \
            "the char_to_idx mapping functions should have been initialized"
        if len(self.char_to_idx_mapping_functions) == 0:
            return df
        df = df.copy()
        for column, mapping_function in self.char_to_idx_mapping_functions.items():
            df[column] = df[column].apply(lambda seq: self._sequence_to_indices(seq,
                                                                                mapping_function,
                                                                                self.char_idx_dtype))
        return df

    @staticmethod
    def stack_numpy_arrays_in_dataframe(df: pd.DataFrame) -> pd.DataFrame:
        return df.apply(lambda item: [np.stack(item)])

    def preprocess_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        """

        :param df:
        :return: a stacked dataframe (with one single row)
        """
        df = self.normalize_columns(df)
        df = self.pad_array_columns(df)
        df = self.sequence_column_to_indices(df)
        df = self.stack_numpy_arrays_in_dataframe(df)
        return df

    def stacked_df_to_dataset(self, stacked_df: pd.DataFrame) -> tf.data.Dataset:
        assert len(stacked_df) == 1, "all column values should be stacked at this point"
        training_data = tuple(stacked_df[self.training_data_columns].iloc[0])
        target_data = tuple(stacked_df[self.target_data_columns].iloc[0])
        tf_dataset = tf.data.Dataset.from_tensor_slices((training_data, target_data))
        return tf_dataset

    def split_dataframe_by_column_values(self, df: pd.DataFrame, tf_dataset_output_file_path: str) \
            -> List[Tuple[str, pd.DataFrame]]:
        if self.split_on_column_values_of is None:
            return [(tf_dataset_output_file_path, df)]
        if len(self.split_on_column_values_of) == 1:
            value_groups = [((values,), df_split) for values, df_split in df.groupby(self.split_on_column_values_of)]
        else:
            value_groups = [(values, df_split) for values, df_split in df.groupby(self.split_on_column_values_of)]
        results = [(
            os.path.join(
                tf_dataset_output_file_path,
                *[str(value).replace("/", "_") for value in values]),
            df_split.drop(columns=self.split_on_column_values_of)
        ) for values, df_split in value_groups]

        return results

    def convert_df_file_to_dataset_file(self,
                                        df_input_file_path: str,
                                        tf_dataset_output_file_path: str) -> None:
        df = pd.read_parquet(df_input_file_path)
        df_splits = self.split_dataframe_by_column_values(df, tf_dataset_output_file_path)
        if len(df_splits) == 0:
            return

        tf_dataset = None

        for path, df_split in df_splits:
            preprocessed_df = self.preprocess_dataframe(df_split)
            tf_dataset = self.stacked_df_to_dataset(preprocessed_df)
            tf.data.experimental.save(dataset=tf_dataset,
                                      path=path,
                                      compression='GZIP')
        assert tf_dataset is not None

        self.logger.debug(tf_dataset.element_spec)

    def __call__(self, item: Tuple[int, str]) -> Optional[str]:
        idx, path = item
        tf_dataset_path = os.path.join(self.dataset_dump_path_prefix, path.split(os.path.sep)[-1])

        info_text = f"Processing item {idx + 1}/{self.item_count}: '{path}'"
        if idx % 10 == 0:
            self.logger.info(info_text)
        else:
            self.logger.debug(info_text)

        if self.skip_existing and os.path.exists(tf_dataset_path):
            self.logger.debug(f"Skipped '{path}' because '{tf_dataset_path}' already exists")
            return None

        self.convert_df_file_to_dataset_file(df_input_file_path=path,
                                             tf_dataset_output_file_path=tf_dataset_path)
        gc.collect()

        return tf_dataset_path

    def process(self, parquet_file_paths: Iterable[str], **kwargs: Dict[str, Any]) -> List[str]:
        item_processor = ItemProcessor(
            items=enumerate(parquet_file_paths),
            item_processor=self.__call__,
            action_name="parquet2tf_dataset-process",
            subject_name="mzmlid parquet file",
            logger=self.logger,
            **kwargs  # type: ignore
        )
        results: List[str] = list(item_processor.process())  # type: ignore
        return results


class DatasetLoader:
    def __init__(
            self,
            element_spec: Iterable[Union[Iterable[tf.TensorSpec], tf.TensorSpec]],
            batch_size: Optional[int] = 32,
            drop_batch_remainder: bool = True,
            shuffle_buffer_size: Optional[int] = 200_000,
            reshuffle_each_iteration: bool = True,
            prefetch_mode: Optional[int] = tf.data.experimental.AUTOTUNE,
            run_benchmarks: bool = False,
            deterministic: Optional[bool] = False,
            thread_count: Optional[int] = os.cpu_count(),
            cache_path: Optional[str] = None,
            keep_cache: bool = True,
            logger: log.Logger = log.DEFAULT_LOGGER,
    ):
        self.element_spec = element_spec
        self.batch_size = batch_size
        self.drop_batch_remainder = drop_batch_remainder
        self.shuffle_buffer_size = shuffle_buffer_size
        self.reshuffle_each_iteration = reshuffle_each_iteration
        self.prefetch_mode = prefetch_mode
        self.run_benchmarks = run_benchmarks
        self.deterministic = deterministic
        if thread_count is None:
            thread_count = 1
        self.thread_count = thread_count
        self.cache_path = cache_path
        self.keep_cache = keep_cache
        self.logger = logger

    def _load_dataset_from_file(self, path: str) -> tf.data.Dataset:
        return tf.data.experimental.load(
                path=path,
                element_spec=self.element_spec,
                compression='GZIP'
            )

    def _load_dataset_interleaved(self, paths: List[str]) -> tf.data.Dataset:
        return tf.data.Dataset.from_tensor_slices(paths).interleave(
            map_func=self._load_dataset_from_file,
            num_parallel_calls=self.thread_count,
            deterministic=self.deterministic,
        )

    def _cache_dataset(self, dataset: tf.data.Dataset, name: str) -> tf.data.Dataset:
        if self.cache_path is None:
            return dataset

        try:
            if not self.keep_cache:
                shutil.rmtree(self.cache_path)
        except FileNotFoundError:
            pass

        utils.ensure_dir_exists(self.cache_path)
        return dataset.cache(os.path.join(self.cache_path, name))

    def _shuffle_dataset(self, dataset: tf.data.Dataset) -> tf.data.Dataset:
        if self.shuffle_buffer_size is None:
            return dataset
        return dataset.shuffle(
            buffer_size=self.shuffle_buffer_size,
            reshuffle_each_iteration=self.reshuffle_each_iteration
        )

    def _batch_dataset(self, dataset: tf.data.Dataset) -> tf.data.Dataset:
        if self.batch_size is None:
            return dataset
        return dataset.batch(
            batch_size=self.batch_size,
            drop_remainder=self.drop_batch_remainder,
            deterministic=self.deterministic,
        )

    def _prefetch_dataset(self, dataset: tf.data.Dataset) -> tf.data.Dataset:
        if self.prefetch_mode is None:
            return dataset
        return dataset.prefetch(
            buffer_size=self.prefetch_mode
        )

    def _run_benchmark(self, dataset: tf.data.Dataset, name: str) -> None:
        if self.run_benchmarks:
            self.logger.info(f"running benchmark for '{name}' dataset")
            tfds.benchmark(dataset)
            gc.collect()
            self.logger.info(f"ran benchmark for '{name}' dataset - waiting 5 seconds")
            time.sleep(5)

    def prepare_dataset(self, paths: Union[str, List[str]], name: str = 'my_dataset') -> tf.data.Dataset:
        if isinstance(paths, str):
            path_list: List[str] = [paths]
        else:
            path_list = paths
        dataset = self._load_dataset_interleaved(path_list)
        dataset = self._shuffle_dataset(dataset)
        dataset = self._batch_dataset(dataset)
        dataset = self._cache_dataset(dataset, name)
        self._run_benchmark(dataset, name)

        return dataset

    def load_datasets_by_type(self, dataset_file_paths: Dict[str, List[str]]) -> Dict[str, tf.data.Dataset]:
        datasets = {
            training_data_type: self.prepare_dataset(paths, name=training_data_type)
            for training_data_type, paths in dataset_file_paths.items()
        }
        return datasets



from typing import Union, Callable

import tensorflow as tf
from tensorflow.python import keras as K


class MaskedLoss(K.losses.LossFunctionWrapper):
    def __init__(
            self,
            loss_function: Callable[[tf.Variable, tf.Variable], tf.Variable],
            masking_value: Union[str, int, float],
            name: str = 'masked_loss',
            reduction: str = tf.keras.losses.Reduction.NONE):
        def _masked_loss(
                y_true: tf.Variable,
                y_pred: tf.Variable,
        ) -> tf.Variable:
            y_true = tf.squeeze(y_true, name="masked_loss__squeezed_y_true")
            y_pred = tf.squeeze(y_pred, name="masked_loss__squeezed_y_pred")

            length_mask = tf.equal(y_true, masking_value,
                                   name="masked_loss__is_masking_value")

            length_mask = tf.cast(length_mask, tf.float32,
                                  name="masked_loss__is_masking_value_float")

            length_mask = tf.math.subtract(
                tf.constant(
                    value=1,
                    dtype=tf.float32
                ), length_mask, name="masked_loss__is_masking_value_inverted")

            lengths = tf.math.reduce_sum(length_mask, axis=-1,
                                         name="masked_loss__sum_to_get_lengths")

            # to also include the first padding character
            lengths = tf.math.add(
                lengths, 1,
                name="masked_loss__sum_to_include_first_padding"
            )

            mask = tf.sequence_mask(
                lengths=lengths,
                maxlen=y_pred.shape[-2],
                # pre-last dimension = padding length
                # last dimension = one-hot-encoded alphabet
                dtype=tf.float32,
                name="masked_loss__create_sequence_mask"
            )

            losses = loss_function(y_true, y_pred)
            losses = tf.math.multiply(
                losses,
                mask,
                name="masked_loss__apply_sequence_mask"
            )

            summed_losses = tf.math.reduce_sum(losses, axis=-1,
                                               name="masked_loss__sum_losses")

            average_losses = tf.math.divide_no_nan(
                summed_losses,
                lengths,
                name="masked_loss__average_losses"
            )

            return average_losses

        super(MaskedLoss, self).__init__(_masked_loss, name=name,
                                         reduction=reduction)



import gc
import os
from typing import Iterable, List, Optional, Tuple, Dict, Any, Union, Callable, Sequence

import pandas as pd
from mmproteo.utils import log, utils
from mmproteo.utils.config import Config
from mmproteo.utils.filters import AbstractFilterConditionNode, filter_files_list
from mmproteo.utils.formats import read
from mmproteo.utils.processing import ItemProcessor
from pyteomics import mzid, mzml


def merge_mzml_and_mzid_dfs(mzml_df: pd.DataFrame,
                            mzid_df: pd.DataFrame,
                            mzml_key_columns: List[str] = None,
                            mzid_key_columns: List[str] = None,
                            logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    if mzml_key_columns is None:
        mzml_key_columns = Config.default_mzml_key_columns
    if mzid_key_columns is None:
        mzid_key_columns = Config.default_mzid_key_columns

    input_length = min(len(mzml_df), len(mzid_df))

    logger.debug("Started merging MzML and MzID dataframes")
    merged_df = mzml_df.merge(right=mzid_df,
                              how='inner',
                              left_on=mzml_key_columns,
                              right_on=mzid_key_columns)
    output_length = len(merged_df)
    unmatched_rows = input_length - output_length
    if unmatched_rows < 0:
        if logger.is_verbose():
            unique_mzml_length = len(mzml_df.drop_duplicates(inplace=False, subset=mzml_key_columns))
            unique_mzid_length = len(mzid_df.drop_duplicates(inplace=False, subset=mzid_key_columns))
            duplicate_mzml_rows = len(mzml_df) - unique_mzml_length
            duplicate_mzid_rows = len(mzid_df) - unique_mzid_length
            logger.debug("Duplicate MzML rows = %d" % duplicate_mzml_rows)
            logger.debug("Duplicate MzID rows = %d" % duplicate_mzid_rows)
        logger.warning("The key columns were no real key columns, because duplicates were found.")

    logger.debug("Finished merging MzML and MzID dataframes and matched %d x %d -> %d rows" %
                 (len(mzml_df), len(mzid_df), output_length))
    return merged_df


def merge_mzml_and_mzid_files(mzml_filename: str,
                              mzid_filename: str,
                              mzml_key_columns: Optional[List[str]] = None,
                              mzid_key_columns: Optional[List[str]] = None,
                              logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    if mzml_key_columns is None:
        mzml_key_columns = Config.default_mzml_key_columns
    if mzid_key_columns is None:
        mzid_key_columns = Config.default_mzid_key_columns

    logger.debug("Started Merge: '%s' + '%s' -> dataframe" % (mzml_filename, mzid_filename))
    mzml_df = read.read(filename=mzml_filename, logger=logger)
    mzid_df = read.read(filename=mzid_filename, logger=logger)

    merged_df = merge_mzml_and_mzid_dfs(mzml_df=mzml_df,
                                        mzid_df=mzid_df,
                                        mzml_key_columns=mzml_key_columns,
                                        mzid_key_columns=mzid_key_columns,
                                        logger=logger)
    logger.debug("Finished Merge: '%s' + '%s' -> dataframe" % (mzml_filename, mzid_filename))
    return merged_df


class Mz2ParquetMergeJobProcessor:
    def __init__(self,
                 merge_job_count: int,
                 skip_existing: bool = Config.default_skip_existing,
                 mzml_key_columns: Optional[List[str]] = None,
                 mzid_key_columns: Optional[List[str]] = None,
                 logger: log.Logger = log.DEFAULT_LOGGER):
        self.merge_job_count = merge_job_count
        self.skip_existing = skip_existing
        self.mzml_key_columns = mzml_key_columns
        self.mzid_key_columns = mzid_key_columns
        self.logger = logger

    def _process_mzml_and_mzid_to_parquet_merge_job(self,
                                                    mzml_filename: str,
                                                    mzid_filename: str,
                                                    target_filename: str,
                                                    current_merge_job_index: int) -> Optional[str]:
        if self.skip_existing and os.path.exists(target_filename):
            self.logger.info("Skipping Merge %d/%d: '%s' + '%s' -> '%s' already exists" %
                             (current_merge_job_index + 1, self.merge_job_count, mzml_filename, mzid_filename,
                              target_filename))
            return None

        self.logger.info("Started Merge %d/%d: '%s' + '%s' -> '%s'" %
                         (current_merge_job_index + 1, self.merge_job_count, mzml_filename, mzid_filename,
                          target_filename))
        merged_df = merge_mzml_and_mzid_files(mzml_filename=mzml_filename,
                                              mzid_filename=mzid_filename,
                                              mzml_key_columns=self.mzml_key_columns,
                                              mzid_key_columns=self.mzid_key_columns,
                                              logger=self.logger)
        merged_df.to_parquet(path=target_filename)
        self.logger.info("Finished Merge %d/%d: '%s' + '%s' -> '%s'" %
                         (current_merge_job_index + 1, self.merge_job_count, mzml_filename, mzid_filename,
                          target_filename))

        return target_filename

    def __call__(self, merge_job: Tuple[int, Tuple[str, str, str]]) -> Optional[str]:
        index, (mzml_filename, mzid_filename, target_filename) = merge_job
        return self._process_mzml_and_mzid_to_parquet_merge_job(mzml_filename=mzml_filename,
                                                                mzid_filename=mzid_filename,
                                                                target_filename=target_filename,
                                                                current_merge_job_index=index)


def _create_merge_jobs(filenames_and_extensions: List[Tuple[str, Tuple[str, str]]],
                       prefix_length_tolerance: int,
                       target_filename_postfix: str,
                       skip_existing: bool = Config.default_skip_existing,
                       logger: log.Logger = log.DEFAULT_LOGGER) -> List[Tuple[str, str, str]]:
    if len(filenames_and_extensions) < 2:
        return list()

    filenames_and_extensions = sorted(filenames_and_extensions)
    merge_jobs = []

    first_filename_and_extension: Tuple[Optional[str], Tuple[str, str]] = filenames_and_extensions[0]
    last_filename, (last_filename_prefix, last_extension) = first_filename_and_extension
    for filename, (filename_prefix, extension) in filenames_and_extensions[1:]:
        next_last_filename: Optional[str] = filename

        if last_filename is not None and extension != last_extension:
            common_filename_prefix_length = len(os.path.commonprefix([filename_prefix, last_filename_prefix]))
            required_filename_prefix_length = min(len(filename_prefix),
                                                  len(last_filename_prefix)) - prefix_length_tolerance
            if common_filename_prefix_length >= required_filename_prefix_length:
                # found a possible merging pair
                if extension == "mzml":
                    mzml_filename = filename
                    mzid_filename = last_filename
                else:
                    mzml_filename = last_filename
                    mzid_filename = filename

                target_filename = filename[:common_filename_prefix_length] + target_filename_postfix
                if skip_existing and os.path.exists(target_filename):
                    logger.info(f"Skipping Merge: '{mzml_filename}' + '{mzid_filename}' "
                                f"-> '{target_filename}' already exists")
                else:
                    merge_jobs.append((mzml_filename, mzid_filename, target_filename))

                next_last_filename = None  # skip next iteration to prevent merging the same file with multiple others
        last_filename = next_last_filename
        last_filename_prefix = filename_prefix
        last_extension = extension

    return merge_jobs


def merge_mzml_and_mzid_files_to_parquet(filenames: Sequence[Optional[str]],
                                         skip_existing: bool = Config.default_skip_existing,
                                         max_num_files: Optional[int] = None,
                                         count_failed_files: bool = Config.default_count_failed_files,
                                         count_skipped_files: bool = Config.default_count_skipped_files,
                                         column_filter: Optional[AbstractFilterConditionNode] = None,
                                         mzml_key_columns: Optional[List[str]] = None,
                                         mzid_key_columns: Optional[List[str]] = None,
                                         prefix_length_tolerance: int = 0,
                                         target_filename_postfix: str = Config.default_mzmlid_parquet_file_postfix,
                                         thread_count: int = Config.default_thread_count,
                                         logger: log.Logger = log.DEFAULT_LOGGER) -> List[str]:
    filenames = filter_files_list(filenames=filenames,
                                  column_filter=column_filter,
                                  sort=False,
                                  logger=logger)

    filenames_and_extensions: List[Tuple[str, Tuple[str, str]]] \
        = [(filename, read.separate_extension(filename=filename, extensions={"mzml", "mzid"}))
           for filename in filenames if filename is not None]
    filenames_and_extensions = [(filename, (file, ext)) for filename, (file, ext) in filenames_and_extensions if
                                len(ext) > 0]
    if len(filenames_and_extensions) < 2:
        logger.warning("No MzML and MzID files available for merging")
        return []

    merge_jobs = _create_merge_jobs(filenames_and_extensions=filenames_and_extensions,
                                    prefix_length_tolerance=prefix_length_tolerance,
                                    target_filename_postfix=target_filename_postfix,
                                    skip_existing=skip_existing,
                                    logger=logger)

    item_processor = Mz2ParquetMergeJobProcessor(merge_job_count=len(merge_jobs),
                                                 skip_existing=skip_existing,
                                                 mzml_key_columns=mzml_key_columns,
                                                 mzid_key_columns=mzid_key_columns,
                                                 logger=logger)

    processing_results: Iterable[Optional[Any]] = ItemProcessor(items=enumerate(merge_jobs),
                                                                item_processor=item_processor.__call__,
                                                                action_name="mzmlid-merge",
                                                                subject_name="file pair",
                                                                max_num_items=max_num_files,
                                                                count_failed_items=count_failed_files,
                                                                count_null_results=count_skipped_files,
                                                                keep_null_values=False,
                                                                thread_count=thread_count,
                                                                logger=logger).process()
    parquet_files: List[str] = list(processing_results)  # type: ignore
    return parquet_files


def _prepare_mzid_entry(entry: Dict[str, Any]) -> Dict[str, Any]:
    entry = entry.copy()

    try:
        items = utils.list_of_dicts_to_dict_by_key(entry["SpectrumIdentificationItem"], "rank")
        if items is not None:
            entry["SpectrumIdentificationItem"] = items
    except Exception:
        pass

    entry = utils.flatten_dict(entry)
    return entry


def read_mzid(filename: str, logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    entries = read.iter_entries(mzid.read(filename), logger=logger)
    extracted_entries = [_prepare_mzid_entry(entry) for entry in entries]
    return pd.DataFrame(data=extracted_entries)


def read_mzml(filename: str, logger: log.Logger = log.DEFAULT_LOGGER) -> pd.DataFrame:
    entries = read.iter_entries(mzml.read(filename), logger=logger)
    extracted_entries = [utils.flatten_dict(entry) for entry in entries]
    return pd.DataFrame(data=extracted_entries)


class FilteringProcessor:
    default_is_decoy_column_name = 'SpectrumIdentificationItem__1__PeptideEvidenceRef__0__isDecoy'
    default_fdr_column_name = 'SpectrumIdentificationItem__1__MSGFQValue'
    default_peptide_sequence_column_name = 'peptide_sequence'
    default_mz_array_column_name = 'mz_array'
    default_intensity_array_column_name = 'intensity_array'

    @staticmethod
    def get_default_output_columns() -> List[str]:
        return [
            FilteringProcessor.default_peptide_sequence_column_name,
            FilteringProcessor.default_mz_array_column_name,
            FilteringProcessor.default_intensity_array_column_name,
        ]

    def __init__(self,
                 dump_path: str,
                 fdr: float = 0.01,
                 skip_existing: bool = True,
                 is_decoy_column_name: str = default_is_decoy_column_name,
                 fdr_column_name: str = default_fdr_column_name,
                 output_columns: Optional[List[str]] = None,
                 post_processor: Callable[[pd.DataFrame], pd.DataFrame] = lambda x: x,
                 logger: log.Logger = log.DEFAULT_LOGGER):
        self.is_decoy_column_name = is_decoy_column_name
        self.fdr_column_name = fdr_column_name
        self.fdr = fdr
        if output_columns is None:
            self.output_columns = self.get_default_output_columns()
        else:
            self.output_columns = output_columns
        self.dump_path = dump_path.rstrip(os.path.sep)
        self.skip_existing = skip_existing
        self.post_processor = post_processor
        self.logger = logger

    def __call__(self, input_file_path: str) -> Optional[Dict[str, Union[str, int, float]]]:
        output_file_path = self.dump_path + os.path.sep + input_file_path.split(os.path.sep)[-1]
        if self.skip_existing and os.path.exists(output_file_path):
            self.logger.debug(f"Skipping filtering '{input_file_path}' -> '{output_file_path}' already exists")
            return None

        res: Dict[str, Union[str, int, float]] = {
            'input_file_path': input_file_path,
            'output_file_path': output_file_path,
        }

        df = pd.read_parquet(input_file_path)
        length = len(df)
        res['original_sequence_count'] = length

        df = df.dropna(subset=[self.is_decoy_column_name])
        new_length = len(df)
        res['NaN_decoy_count'] = length - new_length
        length = new_length

        df = df[df[self.fdr_column_name] <= self.fdr]
        new_length = len(df)
        res['above_fdr_count'] = length - new_length
        length = new_length

        decoy_counts = df[self.is_decoy_column_name].value_counts()
        left_decoys: int = decoy_counts.get(True, 0)
        left_targets: int = decoy_counts.get(False, 0)
        res['left_decoys'] = left_decoys
        res['left_targets'] = left_targets
        res['fdr'] = left_decoys / left_targets

        # filter out Decoys

        df = df[~df[self.is_decoy_column_name].astype(bool)]
        new_length = len(df)
        res['removed_decoys'] = length - new_length
        length = new_length

        df = self.post_processor(df)
        new_length = len(df)
        res['removed_by_post_processor'] = length - new_length
        length = new_length

        res['final_sequence_count'] = length

        df = df[self.output_columns]
        df.to_parquet(output_file_path)

        self.logger.info(f"Finished filtering '{input_file_path}' -> '{output_file_path}'")

        return res


def filter_files(input_file_paths: List[str],
                 output_path: str,
                 fdr: float = 0.01,
                 skip_existing: bool = True,
                 is_decoy_column_name: str = FilteringProcessor.default_is_decoy_column_name,
                 fdr_column_name: str = FilteringProcessor.default_fdr_column_name,
                 output_columns: Optional[List[str]] = None,
                 post_processor: Callable[[pd.DataFrame], pd.DataFrame] = lambda x: x,
                 thread_count: int = Config.default_thread_count,
                 logger: log.Logger = log.DEFAULT_LOGGER) -> List[Dict[str, Union[str, int, float]]]:
    filter_processor = FilteringProcessor(dump_path=output_path,
                                          fdr=fdr,
                                          skip_existing=skip_existing,
                                          is_decoy_column_name=is_decoy_column_name,
                                          fdr_column_name=fdr_column_name,
                                          output_columns=output_columns,
                                          post_processor=post_processor,
                                          logger=logger)
    output_files = list(ItemProcessor(
        items=input_file_paths,
        item_processor=filter_processor.__call__,
        action_name="fdr-filter",
        subject_name="mzmlid file",
        thread_count=thread_count,
        logger=logger
    ).process())
    return output_files  # type: ignore


class MzmlidFileStatsCreator:
    def __init__(
            self,
            mzmlid_file_paths: List[str],
            statistics_file_path: str,
            seq_col_name: str = FilteringProcessor.default_peptide_sequence_column_name,
            int_col_name: str = FilteringProcessor.default_intensity_array_column_name,
            logger: log.Logger = log.DEFAULT_LOGGER
    ):
        self.mzmlid_file_paths = mzmlid_file_paths
        self.file_path_count = len(mzmlid_file_paths)
        self.statistics_file_path = statistics_file_path
        self.SEQ = seq_col_name
        self.INT = int_col_name
        self.logger = logger

    def __call__(self, item: Tuple[int, str]) -> Dict[str, Any]:
        idx, path = item
        info_text = f"Processing item {idx + 1}/{self.file_path_count} '{path}'"
        if idx % 10 == 0:
            self.logger.info(info_text)
        else:
            self.logger.debug(info_text)
        df = pd.read_parquet(path)
        max_sequence_length = df[self.SEQ].str.len().max()
        max_array_length = df[self.INT].str.len().max()
        alphabet = set.union(*df[self.SEQ].apply(set))
        item_count = len(df)
        del df
        gc.collect()

        return {
            "file_path": path,
            "max_sequence_length": max_sequence_length,
            "max_array_length": max_array_length,
            "alphabet": alphabet,
            "item_count": item_count
        }

    def process(self, **kwargs: Dict[str, Any]) -> pd.DataFrame:
        if os.path.exists(self.statistics_file_path):
            file_stats = pd.read_parquet(self.statistics_file_path)
            file_stats.alphabet = file_stats.alphabet.apply(set)
            self.logger.info(f"loaded previous statistics file '{self.statistics_file_path}'")
            return file_stats
        else:
            file_stats = pd.DataFrame(
                ItemProcessor(
                    items=enumerate(self.mzmlid_file_paths),
                    item_processor=self.__call__,
                    action_name="analyse",
                    subject_name="mzmlid file",
                    logger=self.logger,
                    **kwargs  # type: ignore
                ).process()
            )

            file_stats_writable = file_stats.copy()
            file_stats_writable.alphabet = file_stats_writable.alphabet.apply(list)  # cannot store sets
            file_stats_writable.to_parquet(self.statistics_file_path)

            self.logger.info(f"saved statistics to '{self.statistics_file_path}' file")

            return file_stats



import tensorflow as tf

from typing import Union, Tuple, Any, List


class InputLayerConfiguration:
    def __init__(self, name: str, shape: Union[int, Tuple[Any]], mask_value: Any):
        self.name = name
        self.shape = shape
        self.mask_value = mask_value

    def get_shape_tuple(self) -> Tuple[Any]:
        if not isinstance(self.shape, tuple):
            return self.shape,
        return self.shape


def create_masked_input_layers(input_layer_configurations: List[InputLayerConfiguration]) \
        -> Tuple[List[tf.keras.layers.Input], List[tf.keras.layers.Masking]]:
    named_input_layers = {
        config.name: tf.keras.layers.Input(
            shape=config.get_shape_tuple(),
            name=config.name
        ) for config in input_layer_configurations
    }
    masked_input_layers = {
        config.name: tf.keras.layers.Masking(
            mask_value=config.mask_value,
            name=f"masked_{config.name}",
        )(named_input_layers[config.name])
        for config in input_layer_configurations
    }
    named_input_layers_list = [named_input_layers[config.name] for config in input_layer_configurations]
    masked_input_layers_list = [masked_input_layers[config.name] for config in input_layer_configurations]
    return (named_input_layers_list, masked_input_layers_list)



import glob
import os
import random
from typing import Dict, List, Optional, Tuple, Iterable, Set

from mmproteo.utils import visualization, utils, log


def _get_values_at_wildcard_path_position(wildcard_path: str, path_position: int) -> Set[str]:
    return {path.split(os.path.sep)[path_position] for path in glob.glob(wildcard_path)}


def _create_placeholder_path(wildcard_path: str, path_position: int, placeholder_name: str = 'value') -> str:
    parts = wildcard_path.split(os.path.sep)
    parts[path_position] = f"{{{placeholder_name}}}"
    return os.path.sep.join(parts)


def _assign_values_randomly_to_splits(values: Iterable[str], splits: Dict[str, float]) -> Dict[str, List[str]]:
    prefixed_splits: Dict[Optional[str], float] = splits  # type: ignore
    prefixed_splits[None] = 0

    sorted_splits: List[Tuple[Optional[str], float]] = sorted(prefixed_splits.items(), key=lambda tupl: tupl[1])
    assert sorted_splits[0][0] is None, "the added zero should also be the zeroth element"

    shuffled_values = list(values)
    random.shuffle(shuffled_values)

    assigned_values = {
        category: shuffled_values[
                            int(sorted_splits[i][1] * len(shuffled_values)):
                            int(split * len(shuffled_values))
                            ]
        for i, (category, split) in enumerate(sorted_splits[1:])
    }
    return assigned_values  # type: ignore


def _find_file_paths_per_category_value(
        assigned_values: Dict[str, List[str]],
        wildcard_path_with_placeholder: str,
        placeholder_name: str = 'value'
) -> Dict[str, List[str]]:
    return {
        category: utils.flatten_single_list(
            [
                glob.glob(wildcard_path_with_placeholder.format(**{placeholder_name: value})) for value in values
            ]
        ) for category, values in assigned_values.items()
    }


def _store_assigned_file_paths_as_json(
        assigned_file_paths: Dict[str, List[str]],
        output_file: str) -> str:
    with open(output_file, 'w') as file:
        file.write(visualization.pretty_print_json(assigned_file_paths))

    return output_file


def assign_wildcard_paths_to_splits_grouped_by_path_position_value(
        wildcard_path: str,
        path_position: int,
        splits: Dict[str, float],
        paths_dump_file: Optional[str] = None,
        skip_existing: bool = True,
        logger: log.Logger = log.DEFAULT_LOGGER
) -> Dict[str, List[str]]:
    """
    Collect and assign paths from a given wildcard path and split the paths grouped by values at a given path
    position.

    :param wildcard_path:   a path containing '*'s to match several files/directories simultaneously
    :param path_position:   a path-separator separated position index in the given wildcard path. Negative index
                            values matches the path snippets count reversely
    :param splits:          a dictionary containing the categories, to which the paths should be assigned, as keys.
                            The corresponding float values represent the upper bound of the share for this category.
                            The float values should be accumulative.
    :param paths_dump_file: a file path to use for storing/loading the (random) path assignment (optional)
    :param skip_existing:   whether to try loading path assignments from
                        :paramref:`split_collected_paths_by_path_position_value.paths_dump_file`
    :param logger:          the logger instance to use
    :return: a dictionary with the category values of the splits as keys and lists of assigned file paths as values
    """
    if skip_existing and paths_dump_file is not None and os.path.exists(paths_dump_file):
        file_paths = utils.load_json(paths_dump_file)
        logger.info(f"found file paths dump '{paths_dump_file}'")
        return file_paths

    categorical_values = _get_values_at_wildcard_path_position(wildcard_path, path_position)
    assigned_values = _assign_values_randomly_to_splits(categorical_values, splits=splits)
    logger.debug("assigned values:")
    visualization.print_list_length_in_dict(assigned_values, print_func=logger.debug)

    file_paths = _find_file_paths_per_category_value(
        assigned_values=assigned_values,
        wildcard_path_with_placeholder=_create_placeholder_path(
            wildcard_path=wildcard_path,
            path_position=path_position,
            placeholder_name='value'
        ),
        placeholder_name='value',
    )
    if paths_dump_file is not None:
        _store_assigned_file_paths_as_json(file_paths, paths_dump_file)
        logger.info(f"dumped file paths into '{paths_dump_file}'")
    return file_paths



import datetime
import os
import shutil
from typing import Union, List

import tensorflow as tf
from tqdm.keras import TqdmCallback


def create_tensorboard_callback(
        tensorboard_log_dir: str = "logs",
        training_type: str = "fit",
        keep_logs: bool = True,
) -> tf.keras.callbacks.TensorBoard:
    if not keep_logs:
        try:
            shutil.rmtree(tensorboard_log_dir)
        except FileNotFoundError:
            pass
    return tf.keras.callbacks.TensorBoard(
        log_dir=os.path.join(
            tensorboard_log_dir,
            training_type,
            datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        ),
        histogram_freq=1,
    )


# Parts of the following code are based on code by Sven Giese:
# https://github.com/Rappsilber-Laboratory/xiRT/blob/master/xirt/xirtnet.py
# (28.06.21; method 'xiRTNET.get_callbacks')
# https://doi.org/10.5281/zenodo.4669957

def create_progressbar_callback(verbose: int = 0) -> TqdmCallback:
    return TqdmCallback(verbose=verbose)


def create_reduce_learning_rate_callback(
        monitor: str = 'val_loss',
        reduce_lr_factor: float = 0.1,
        reduce_lr_patience: int = 10,
        verbose: int = 1,
        min_delta: float = 1e-4,
        mode: str = 'min',
) -> tf.keras.callbacks.ReduceLROnPlateau:
    return tf.keras.callbacks.ReduceLROnPlateau(
        monitor=monitor,
        factor=reduce_lr_factor,
        patience=reduce_lr_patience,
        verbose=verbose,
        min_delta=min_delta,
        mode=mode
    )


def create_early_stopping_callback(
        monitor: str = 'val_loss',
        mode: str = 'min',
        verbose: int = 1,
        patience: int = 10,
        restore_best_weights: bool = True,
) -> tf.keras.callbacks.EarlyStopping:
    return tf.keras.callbacks.EarlyStopping(
        monitor=monitor,
        mode=mode,
        verbose=verbose,
        patience=patience,
        restore_best_weights=restore_best_weights
    )


def create_checkpoint_callback(
        filepath: str,
) -> tf.keras.callbacks.ModelCheckpoint:
    return tf.keras.callbacks.ModelCheckpoint(
        filepath=filepath,
    )


def create_log_csv_callback(
        filename: str
) -> tf.keras.callbacks.CSVLogger:
    return tf.keras.callbacks.CSVLogger(
        filename=filename,
    )


def create_callbacks(
        tensorboard: Union[bool, tf.keras.callbacks.TensorBoard] = False,
        progressbar: Union[bool, TqdmCallback] = False,
        reduce_lr: Union[bool, tf.keras.callbacks.ReduceLROnPlateau] = False,
        early_stopping: Union[bool, tf.keras.callbacks.EarlyStopping] = False,
        checkpoints: Union[bool, tf.keras.callbacks.ModelCheckpoint] = False,
        csv: Union[bool, tf.keras.callbacks.CSVLogger] = False,
        base_path: str = ".",
) -> List[tf.keras.callbacks.Callback]:
    callbacks: List[tf.keras.callbacks.Callback] = []

    if tensorboard is True:
        tensorboard = create_tensorboard_callback(
            os.path.join(base_path, 'tensorboard')
        )
    if tensorboard:
        callbacks.append(tensorboard)

    if progressbar is True:
        progressbar = create_progressbar_callback()
    if progressbar:
        callbacks.append(progressbar)

    if reduce_lr is True:
        reduce_lr = create_reduce_learning_rate_callback()
    if reduce_lr:
        callbacks.append(reduce_lr)

    if early_stopping is True:
        early_stopping = create_early_stopping_callback()
    if early_stopping:
        callbacks.append(early_stopping)

    if checkpoints is True:
        checkpoints = create_checkpoint_callback(
            filepath=os.path.join(base_path, 'checkpoints')
        )
    if checkpoints:
        callbacks.append(checkpoints)

    if csv is True:
        csv = create_log_csv_callback(
            filename=os.path.join(base_path, 'log.csv')
        )
    if csv:
        callbacks.append(csv)

    return callbacks
